/*! \file dos.c \brief DOS-Functions */
//###########################################################
///	\ingroup singlefat
///	\defgroup DOS DOS-Functions (dos.c)
///	\code #include "dos.h" \endcode
///	\par Uebersicht
//###########################################################
// Benutzung auf eigene Gefahr !
//
// Use at your own risk !
//
// If you like it fast compile with -O2
// If you like it small compile with -Os
//
// 28.06.2007 Fseek() from Alex ???.
//            Seeking is only available for files open for reading !
//
// 10.10.2006 U16 tmp1 in Fwrite() and Fread() speeds up !
//            "% BYTE_PER_SEC" replaced with "&= (BYTE_PER_SEC -1)
//            "/ BYTE_PER_SEC" replaced with " >> 9"
//
//#########################################################################
// Last change: 03.08.2007
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################
// Compiler: WinArm 4.1
//#########################################################################
//@{
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "dos.h"

#if defined (FAT_DEBUG_RW_HITS)
 #include "rprintf.h" //for testing only
#endif

//use some global variables for file access
#ifdef USE_FAT32
 U32 FileFirstCluster=0;     //first cluster of file
 U32 FileCurrentCluster=0;   //actual cluster in use
#else
 U16 FileFirstCluster=0;     //first cluster of file
 U16 FileCurrentCluster=0;   //actual cluster in use
#endif

U32 FileCurrentSector=0;    //sector with last data read/written
U8 FileClusterSectorOffset=0; //sector of cluster in use
U32 FileClusterCount=0;     //clusters read (not really)
U32 FileDirSector=0;        //sector with dir entry of file
U8 FileDirOffset=0;        //offset to file entry in FileDirSector
U32 FileSize=0;
U32 FilePosition=0;         //actual position when reading file         
U8 FileFlag=0;		    //read or write
U8 FileAttr=0;		    //file attribute
char FileName[8];
char FileExt[3];


#ifdef DOS_WRITE
#ifdef DOS_DELETE
//###########################################################
/*!\brief Delete a file
 * \param		name	DOS 8.3 name of the file 
 * \return 		F_OK if file could be deleted, F_ERROR if not
 */
unsigned char Remove(char *name)
//###########################################################
{
#ifdef USE_FAT32
 U32 tmp,tmp1;
#else
 U16 tmp,tmp1;
#endif

 if(FileFlag!=0) return F_ERROR; // don't delete if a file is open

 FileAttr=0xFF; // test if we really find a file and not a directory
                // FileAttr then must be ATTR_FILE
                
 if(FindName(name)==FULL_MATCH) // Look if file exists
  {
   if(FileAttr!=ATTR_FILE) return F_ERROR; // this was not a file !

   tmp=FileFirstCluster;

#ifdef STRICT_FILESYSTEM_CHECKING
// Bug found by Michele Ribaudo
   if(FileFirstCluster>=2) // Zero length files made by Win have no cluster chain !
    {
#endif

     // free clusters in FAT cluster chain (make zero)
     do
      {
       tmp1=GetNextClusterNumber(tmp); // save next cluster number
       WriteClusterNumber(tmp,0);      // free cluster
       tmp=tmp1;                       // restore next cluster number
      }while(tmp<endofclusterchain);

#ifdef STRICT_FILESYSTEM_CHECKING
    }
#endif

   if(FATStatus>0)
    {
#ifdef USE_FATBUFFER
     WriteFATSector(FATCurrentSector,fatbuf); // write the FAT buffer
#else //added 22.10.2006
     WriteFATSector(FATCurrentSector,iob); // write the FAT buffer
#endif
     FATStatus=0;
    } 

   FileName[0]=0xE5;   //mark file as deleted. does not affect long filename entrys !
   FileSize=0;         //make filesize 0
   FileFirstCluster=0; //delete first cluster
   FileAttr=0;         //is this necessary ? 
   UpdateFileEntry();
  }
  
 return F_OK;  
}
#endif //DOS_DELETE
#endif //DOS_WRITE

#ifdef DOS_WRITE
//###########################################################
// write one byte to file
// returns 0 if nothing written
// returns 1 if byte is written
//
// Fputc() does not write to CF until a sector is completely
// filled. you can force writing with Fflush() when file should
// keep open, or close the file with Fclose().
//
//###########################################################
/*
U16 Fputc(unsigned char buf)
{
 return Fwrite(&buf,1); //the easy way :)
//maybe its faster to return unsigned char
}
*/

//###########################################################
/*!\brief Write to a file
 * \param		buf	Buffer containing your data
 * \param		count	Number of bytes to be written
 * \return 		Number of bytes written to the file
 *
 * If number of bytes written is 0, your disk is full.

 Fwrite() does not write to CF until a sector is completely
 filled. You can force writing with Fflush() when file should
 keep open, or close the file with Fclose().

 */
//###########################################################
 U16 Fwrite(unsigned char *buf, U16 count)
//###########################################################
{
 unsigned char secoffset;
 U32 tmp;
 U16 tmp1;
 U16 remaining;
 
 U16 buffoffset;
 U16 bytecount;

 if(FileFlag!=F_WRITE) return 0; //don't write if file is closed
                                 //or open for reading !
                                 //written bytes are zero

 if(count==0) return 0; //nothing to write
 
 bytecount=0;
 remaining = count;
 
 do
  {
   tmp=FileSize;         //next write position
   tmp-= FileClusterCount; //calc byte position in cluster
   
   if(tmp >= BytesPerCluster)//is position in new cluster ?
    {
     WriteFileSector(FileCurrentSector,iob); //write last sector
     FileCurrentCluster=AllocCluster(FileCurrentCluster); //alloc new cluster
     if(FileCurrentCluster==DISK_FULL)
      {
       return bytecount; //return number of bytes written before disk full
      } 

     #ifndef USE_FATBUFFER // keep it here holgi !?
      iob_status=IOB_DATA; // we have to track this for FClose() and Fflush() !
     #endif

     FileCurrentSector=GetFirstSectorOfCluster(FileCurrentCluster); //set new 1st sector of cluster
     FileClusterSectorOffset = 0;

//       FileClusterCount++;                   //update cluster count
     FileClusterCount+=BytesPerCluster;   //update cluster count
     tmp-= BytesPerCluster;               //calc new byte position in cluster
    }

   tmp1=(U16)tmp; //we lose upper bits here, but it does not matter !
   
//   buffoffset=(unsigned int)(tmp % BYTE_PER_SEC); //calc offset in sector
   buffoffset = tmp1 & (BYTE_PER_SEC - 1); // tmp % 512 => tmp & (512-1)

//   secoffset=(unsigned char)(tmp1 / BYTE_PER_SEC);  //calc offset from first sector in cluster
   secoffset=(unsigned char)(tmp1 >> 9 );  //calc offset from first sector in cluster

   if(FileClusterSectorOffset != secoffset)
    {
     WriteFileSector(FileCurrentSector,iob); //write last sector used

     FileCurrentSector -= FileClusterSectorOffset;
     FileCurrentSector += secoffset;
     FileClusterSectorOffset = secoffset;
    }
      
  if((buffoffset + remaining) <= BYTE_PER_SEC) // all bytes can be copied to current sector buffer
   {
    tmp = remaining;
   }
  else //all bytes can NOT be copied to current sector buffer
   {
    tmp = BYTE_PER_SEC-buffoffset;
    remaining -= tmp;
   } 

  bytecount += tmp;
  FileSize += tmp; //update Filesize

  U8 *p;
  p = &iob[buffoffset];
  while(tmp--) *p++ = *buf++;

 }while(bytecount<count);

 return bytecount;
}
#endif //DOS_WRITE

#ifdef DOS_READ
//###########################################################
// read one byte from file
// returns the byte
// NO ERROR CHECKING ! Use FileSize to see if you have
// reached the end of the file !
//###########################################################
/*
unsigned char Fgetc(void)
{
 unsigned char by;

 Fread(&by,1);

 return by;
}
*/

//###########################################################
/*!\brief Read from a file
 * \param		buf	Buffer where your data should be stored
 * \param		count	Number of bytes to be read
 * \return 		Number of bytes read from the file
 *
 * If number of bytes read is smaller as count or equal 0, end of file is reached.
 */
 U16 Fread(unsigned char *buf, U16 count)
//###########################################################
{
 unsigned char secoffset;
 U32 tmp;
 U16 tmp1;
 U16 remaining;

 U16 buffoffset;
 U16 bytecount;

 if(FileFlag!=F_READ) return 0; // don't read if file is closed
                                // or open for writing !
                                // read bytes are zero

 if(count==0) return 0; //nothing to read

 if(FilePosition == FileSize) return 0; // don't read after end of file. Also matches FileSize==0
 
 bytecount=0;
 remaining=count;

 tmp = FileSize - FilePosition; // get bytes til end of file
 if(tmp < remaining) remaining = tmp; // smaller as we want to read, then read til end of file
 
 do
  {
   tmp=FilePosition;
//     tmp-= FileClusterCount * BytesPerCluster; //calc byte position in cluster
   tmp-= FileClusterCount; //calc byte position in cluster
//maybe its faster to count cluster and sectorbytes

     if(tmp >= BytesPerCluster)//is position in current cluster ?
      {
       FileCurrentCluster=GetNextClusterNumber(FileCurrentCluster); //if not get next cluster
       FileCurrentSector=GetFirstSectorOfCluster(FileCurrentCluster);//set new 1st sector of cluster
       ReadFileSector(FileCurrentSector,iob); //read new sector
       FileClusterSectorOffset = 0;

//       FileClusterCount++;                   //update cluster count
       FileClusterCount+=BytesPerCluster;   //update cluster count
       tmp-= BytesPerCluster;               //calc new byte position in cluster
      }

     tmp1=(U16)tmp; //we lose upper bits here, but it does not matter !

//   buffoffset=(unsigned int)(tmp % BYTE_PER_SEC); //calc offset in sector
     buffoffset = tmp1 & (BYTE_PER_SEC - 1); // tmp % 512 => tmp & (512-1)

//     secoffset=(unsigned char)(tmp1 / BYTE_PER_SEC);  //calc sector offset from first sector in cluster
     secoffset=(unsigned char)(tmp1 >> 9);  //calc sector offset from first sector in cluster

   if(FileClusterSectorOffset != secoffset) //new sector ?
      {
       FileCurrentSector -= FileClusterSectorOffset;
       FileCurrentSector += secoffset;
       FileClusterSectorOffset = secoffset;
       ReadFileSector(FileCurrentSector,iob); //read new sector
      }
      
    if((buffoffset + remaining) <= BYTE_PER_SEC) // All bytes can be copied from current sector buffer
     {
      tmp = remaining;
     }
    else // all bytes can NOT be copied from current sector buffer
     {
       tmp = BYTE_PER_SEC - buffoffset;
       remaining -= tmp;
     }

    bytecount += tmp;
    FilePosition += tmp;

    U8 *p;
    
    p = &iob[buffoffset];
    while(tmp--) *buf++ = *p++;

  } while(bytecount<count);
  
 return bytecount;
}
#endif //DOS_READ

#if defined (DOS_READ) || defined (DOS_WRITE)
//###########################################################
/*!\brief Open a file
 * \param		name	8.3 DOS name
 * \param		flag	F_READ or F_WRITE
 * \return 		F_OK if successfull, F_ERROR otherwise
 *
 * Open a file for reading OR writing, NOT both !
 */
unsigned char Fopen(char *name, unsigned char flag)
//###########################################################
{
#ifdef DOS_WRITE
 U32 tmp;
#endif //DOS_WRITE


 if(FileFlag!=0) return F_ERROR; //a file is already open !

 FileClusterCount=0;

 #ifdef FAT_DEBUG_RW_HITS
  FATWRHits=0;
  FATRDHits=0;
 #endif
 
#ifdef DOS_READ
 if(flag==F_READ)
  {
   if(FindName(name)==FULL_MATCH) //file MUST exist for reading
    {
#ifdef STRICT_FILESYSTEM_CHECKING
// Bug found by Michele Ribaudo
     if(FileFirstCluster<2) // no clusters allocated for the file !
      {
       FileFlag=0;
       return F_ERROR; // nothing to read from this file
      } 
#endif

     FilePosition=0; //actual read position
     FileCurrentCluster=FileFirstCluster;
     FileFlag=flag; //needed for fclose

     FileCurrentSector=GetFirstSectorOfCluster(FileFirstCluster);
     FileClusterSectorOffset = 0;

     ReadFileSector(FileCurrentSector,iob); //read first sector of file

     return F_OK;      //give back something above zero
    }
  }
#endif //DOS_READ

#ifdef DOS_WRITE
 if(flag==F_WRITE)
  {
   //if file exists, open it and spool to end of file to append data
   if(FindName(name)==FULL_MATCH)
    {
#ifdef STRICT_FILESYSTEM_CHECKING
// Bug found by Michele Ribaudo
     if(FileFirstCluster<2) // no clusters allocated for the file !
      {
        // todo: allocate a cluster
        // You can open this file if you do the following:
        // Remove() the file ! Make a new one with Fopen()
        FileFlag=0; //needed for fclose
        return F_ERROR;
      }
#endif

     tmp=FileFirstCluster;
     FileCurrentCluster=FileFirstCluster; //we need this if file is smaller as ONE cluster
     
     while(tmp<endofclusterchain) //go to end of cluster chain
      {
       tmp=GetNextClusterNumber(tmp);
       if(tmp<endofclusterchain)
        {
         FileCurrentCluster=tmp;
         FileClusterCount+=BytesPerCluster;
        } 
      }

     tmp= FileSize - FileClusterCount;
     FileCurrentSector = GetFirstSectorOfCluster(FileCurrentCluster); 

//     FileClusterSectorOffset = (U8)(tmp / BYTE_PER_SEC);
     FileClusterSectorOffset = (U8)(tmp >> 9);
     FileCurrentSector += FileClusterSectorOffset;

     ReadFileSector(FileCurrentSector,iob); //read first sector of file
    }
   else //make a new file
    { 
     FileAttr=ATTR_FILE; //needed for MakeNewFileEntry() ! 18.10.2006
     MakeFileName(name,FileName,FileExt); //Split into name and extension
     if(MakeNewFileEntry()) //file does not exist, try to make new file in currentdir
      {
       FileCurrentCluster=FileFirstCluster;
       FileCurrentSector = GetFirstSectorOfCluster(FileFirstCluster); 
       FileClusterSectorOffset = 0;

       ReadFileSector(FileCurrentSector,iob); //read first sector of file
      }
     else
      {
       FileFlag=0; //needed for fclose
       return F_ERROR; //new file could not be made
      }
    }

   FileFlag=flag; //needed for fclose

   return F_OK; //file is open for writing
  }//if(flag==F_WRITE)
#endif //DOS_WRITE
  
 return F_ERROR; //something went wrong  
}

//###########################################################
/*!\brief Close a file
 * \return 		Nothing
 */
void Fclose(void)
//###########################################################
{
#ifdef DOS_READ
// nothing todo
// if(FileFlag==F_READ)
//  {
//  }
#endif //DOS_READ

#ifdef DOS_WRITE
 if(FileFlag==F_WRITE)
  {
   Fflush();  //write last sector used to CF and update filesize, filetime

   #ifdef FAT_DEBUG_RW_HITS
    #ifdef USE_FAT32
     printf("FAT WR Hits %lu ",FATWRHits);
    #else
     printf("FAT WR Hits %u ",FATWRHits);
    #endif
   #endif //#ifdef FAT_DEBUG_RW_HITS
  }


#ifdef USE_FATBUFFER
#else
 iob_status=0;
#endif

#endif //DOS_WRITE

 #ifdef FAT_DEBUG_RW_HITS
  #ifdef USE_FAT32
   printf("FAT RD Hits %lu\n",FATRDHits);
  #else
   printf("FAT RD Hits %u\n",FATRDHits);
  #endif
 #endif //#ifdef FAT_DEBUG_RW_HITS

 FileFlag=0; //a second fclose should do nothing
             //reading and writing disabled

}
#endif //#if defined (DOS_READ) || defined (DOS_WRITE)

#ifdef DOS_WRITE
//###########################################################
/*!\brief Flush a file buffer
 * \return 		Nothing
 *
 Force writing last data written into sectorbuffer to be
 stored into CF without Fclose(). Direntry will also be updated.
 If you use a FAT buffer, it will also be written.
 */
void Fflush(void)
//###########################################################
{
 if(FileFlag==0) return; //don't write if file is closed

#ifdef USE_FATBUFFER
 if(FATStatus>0)
  {
   WriteFATSector(FATCurrentSector,fatbuf); // write the FAT buffer
   #ifdef FAT_DEBUG_RW_HITS
    FATWRHits++;
   #endif
   FATStatus=0;
  } 

 WriteFileSector(FileCurrentSector,iob); //write last sector used
#else
 if(iob_status==IOB_FAT && FATStatus>0)
  {
   WriteFATSector(FATCurrentSector,iob); // write the old FAT buffer
   #ifdef FAT_DEBUG_RW_HITS
    FATWRHits++;
   #endif
   FATStatus=0;
  }

 //if last access to iob was a FAT access than we should not write FileCurrentSector here
 if(iob_status==IOB_DATA)
  {
   WriteFileSector(FileCurrentSector,iob); //write last sector used
  }
#endif

 //update file entry filesize and date/time
 UpdateFileEntry();
 ReadFileSector(FileCurrentSector,iob); //iob was changed in UpdateFileEntry()
}
#endif //DOS_WRITE

//############################################################
/*!\brief Search for a filename in current directory
 * \param		name	8.3 DOS name
 * \return 		FULL_MATCH if file exists, NO_MATCH if not
 *
 */
unsigned char FindName(char *name)
//############################################################
{
 unsigned char result;

 if(FileFlag>0) return NO_MATCH; //don't search if a file is open
 
 if(FirstDirCluster<2)
  {
   result=ScanRootDir(name);
  }
 else
  {
   result=ScanSubDir(FirstDirCluster,name);
  }

 return result;
}

#ifdef DOS_FSEEK
/*****************************************************************************/
// First version by Alex ??? 28.06.2007
/*!\brief Seek in a file
 * \param		offset	position to seek to (forward,backward)
 * \param		mode	start position from where to seek
 * \return 		0 if seeking successfull, >0 if not
 *
  Fseek() can (til now) only be used for files open for reading !
  You can not seek above FileSize.
  
  Parameters for mode:

  SEEK_SET	Seek from beginning of file. offset has to be positive.

  SEEK_CUR	Seek from current file pointer position. offset may be positive or negative.

  SEEK_END	Seek from end of file. offset has to be negative.
 */
/*****************************************************************************/
U8 Fseek(S32 offset, U8 mode)
{
 U32 numClusters, curCluster, numSector;
 
 if(FileFlag==0) return 1; //don't seek if file is closed
 if(FileFlag==F_WRITE) return 2; //don't seek if file is open for writing

 // calculate the new File Position
 switch (mode)
  {
   case SEEK_SET: 	
        if(offset<0) return 3; 		// don't seek below 0
        if(offset > FileSize) return 4;   // don't seek above FileSize

//printf("SEEK_SET\n");
        FilePosition = offset;           	
    break;

   case SEEK_END: // offset has to be 0 or negative	!
        if(offset>0) return 5; 		// don't seek above FileSize
        if((FileSize + offset) < 0) return 6; 		// don't seek below 0

//printf("SEEK_END\n");
        FilePosition = FileSize + offset; // offset is negativ !	
    break;

   case SEEK_CUR: 	
        if((FilePosition + offset) < 0) return 7;	// don't seek below 0
        if((FilePosition + offset) > FileSize) return 8;	// don't seek above FileSize

//printf("SEEK_CUR\n");
        FilePosition = FilePosition + offset;	
    break;

   default:
//printf("Unknown SEEK\n");
    return 9;
  }
	
 // test if new FilePosition exists
// if(FilePosition > FileSize) return F_ERROR; // Notbremse
	
 // calculate the current file cluster
 curCluster=FileFirstCluster;
 FileCurrentCluster=FileFirstCluster;
 numClusters=FilePosition/(BYTE_PER_SEC*secPerCluster);
 numSector=(FilePosition/BYTE_PER_SEC)%secPerCluster;
 
 // calculate the cluster address of the new Position (verkettete pointerliste)
 FileClusterCount=0;
 while (numClusters > 0)
  {
   curCluster=GetNextClusterNumber(curCluster);
   numClusters--;
   FileClusterCount+=BytesPerCluster; //update cluster count
  }
    
 //set Current Cluster
 FileCurrentCluster=curCluster;

 // calculate the Sector address of the new Position
 FileCurrentSector = GetFirstSectorOfCluster(curCluster);
 FileClusterSectorOffset = numSector;
 FileCurrentSector += numSector;

 ReadFileSector(FileCurrentSector,iob); //read current sector of file
	
 return 0;
}
#endif //DOS_FSEEK

//@}

