//#########################################################################
// File: dos.h
//
//
// Benutzung auf eigene Gefahr !
//
// Use at your own risk !
//
// 25.07.2007 Added FASTER_FREAD, FASTER_FWRITE
//
// 18.02.2007 Changed F_READ to 'r', F_WRITE to 'w'
//            Now you can do Fopen("mydata.txt",'w');    ;)
//
//#########################################################################
// Last change: 28.07.2007
//#########################################################################
// Compiler: WinARM 4.1.1
//#########################################################################
//@{

#ifndef __DOS_H
#define __DOS_H

#include "typedefs.h" // ARM specific types

// Some defines for more speed optimizations 

// Some more defines for special filesystem handling

#define STRICT_FILESYSTEM_CHECKING // If you define this, some very rare special cases will be
				    // noticed. But this needs more code.
				    // Special cases when files where made on a Win system and
				    // copied to the flash card:
				    //
                                    // Bug found by Michele Ribaudo
				    // Files with zero filesize have no clusters allocated !
				    // Calling Remove() and Fopen() may hang up the system.
				    // If you define STRICT_FILESYSTEM_CHECKING opening a
				    // zero length file for reading or writing gives an error only.
				    // You can remove a zero length file !  
				    //
				    // You don't need this define if all files where made with
				    // my FAT system. Even if filesize is zero.
				    // If unsure keep it defined !

//fopen flags
#define F_CLOSED	0
#define F_READ		'r'  // vorher 1
#define F_WRITE		'w'  // vorher 2

#define F_ERROR		0 // dir/file operation failed
#define F_OK		1 // dir/file operation successfull

#ifndef SEEK_SET
 #define	SEEK_SET	0	/* set file offset to offset */
#endif
#ifndef SEEK_CUR
 #define	SEEK_CUR	1	/* set file offset to current plus offset */
#endif
#ifndef SEEK_END
 #define	SEEK_END	2	/* set file offset to EOF plus offset */
#endif

// #undef defines below in "dosdefs.h" if you don't need them
// spare program memory by deciding if you want to read, write or both
#define DOS_READ	//define this if you want to read files
#define DOS_WRITE	//define this if you want to write files
#define DOS_DELETE	//define this if you want to delete files
#define DOS_READ_RAW	//define this if you want to read files with ReadFileRaw()

#define DOS_CHDIR	//define this if you want to go into subdirectories
#define DOS_MKDIR	//define this if you want to make subdirectories
#define DOS_RENAME	//define this if you want to rename files
                        //DOS_DELETE has to be defined too !
#define DOS_FSEEK       //define this if you want to seek in files (reading only)

// spare program memory by deciding if we want to use FAT12, FAT16, FAT32.
// don't touch if you don't know the FAT type of your drive !
#define USE_FAT12	//define this if you want to use FAT12
#define USE_FAT16	//define this if you want to use FAT16
#define USE_FAT32	//define this if you want to use FAT32

#define USE_FATBUFFER	//define this if you want to use a FAT buffer
                        //needs 517 Bytes of RAM ! 

#define USE_FINDFILE	//define this if you want to use findfirst(); findnext();
#define USE_FINDLONG    //define this if you want to get long filenames
			//from findfirst(); findnext();

#define USE_DRIVEFREE   //define this if you want to get free and used space of your drive

#include "dosdefs.h"  // keep the line at this place ! don't move down or delete 

extern unsigned char Fopen(char *name, unsigned char flag);
extern void Fclose(void);

#ifdef DOS_READ
#ifdef FREAD_SMALL_BUFFERS
 extern unsigned char Fread(unsigned char *buf, unsigned char count);
#else
 extern U16 Fread(unsigned char *buf, U16 count);
#endif
extern unsigned char Fgetc(void);
#endif

#ifdef DOS_WRITE
#ifdef FWRITE_SMALL_BUFFERS
 extern unsigned char Fwrite(unsigned char *buf, unsigned char count);
#else
 extern U16 Fwrite(unsigned char *buf, U16 count);
#endif
extern U16 Fputc(unsigned char buf);
#endif

extern void Fflush(void);
extern unsigned char Remove(char *name);

extern unsigned char ReadFileRaw(char *name);
extern unsigned char FindName(char *name);
extern U8 Rename(char *OldName, char *NewName);

// ###Seek Funktion
extern U8 Fseek(S32 offset, U8 mode);

#define Filelength()	FileSize

#ifdef USE_FAT32
 extern U32 FileFirstCluster;
 extern U32 FileCurrentCluster;
#else
 extern U16 FileFirstCluster;
 extern U16 FileCurrentCluster;
#endif

extern U32 FileCurrentSector;
extern U8 FileClusterSectorOffset;
extern U32 FileClusterCount;
extern U32 FileDirSector;
extern U8 FileDirOffset;
extern U32 FileSize;
extern U32 FilePosition;
extern U8 FileFlag;
extern U8 FileAttr;
extern char FileName[];
extern char FileExt[];

//this is for easier and faster converting from byte arrays to UINT, ULONG
//ui and ul share the same memory space
union Convert {
 U16  ui;
 U32 ul;
};

#ifdef COMPACTFLASH_CARD
 #include "compact.h"
#endif

#ifdef MMC_CARD_SPI
 #include "mmc_spi.h"
#endif

#include "fat.h"
#include "dir.h"

#ifdef USE_FINDFILE
 #include "find_x.h"
#endif

#ifdef USE_DRIVEFREE
 #include "drivefree.h"
#endif

#endif //__DOS_H
//@}
